/*******************************************************************************
 * Copyright 2007  Hon Hai Precision Ind. Co. Ltd.
 * All Rights Reserved.
 * No portions of this material shall be reproduced in any form without the
 * written permission of Hon Hai Precision Ind. Co. Ltd.
 *
 * All information contained in this document is Hon Hai Precision Ind.
 * Co. Ltd. company private, proprietary, and trade secret property and
 * are protected by international intellectual property laws and treaties.
 *
 ******************************************************************************/


#ifdef INCLUDE_QOS

#undef _DEBUG_QOS_INPUT
#undef _DEBUG_QOS_OUTPUT
#undef DEBUG

#include <net/sock.h>
#include "bcmnvram.h"

#define ETHER_TYPE_BCM_TAG      0x8874
#define BCM_TAG_SIZE            6
#define SKB_PORT_OFFSET         17
#define CRC_SIZE                4
#define MIN_PACKET_SIZE         64

static int qos_enable = 0;
static int qos_do_add_strip_tag = 0;

static int igmp_snooping_enable = 0;

extern int dma_qos_enable;      /* flag to tell DMA when QoS is enabled */

#if defined(__CONFIG_IGMP_SNOOPING__)
extern int bcm_robo_snoop_mac_match(uint8 *p_mh_mac);
extern int is_reg_snooping_enable;
#endif

static int et_qos_init(void)
{
    char qos_port[512];
    static int first_time = 1;

    /* Init only once */
    if (!first_time)
        return 0;

    first_time = 0;

    /* Enable / disable QoS */
    if (nvram_match("qos_enable", "1") && 
        !nvram_match("wla_repeater", "1"))
    {
        qos_enable = 1;
    }
    else
    {
        qos_enable = 0;
    }

    strcpy(qos_port, nvram_safe_get("qos_port"));

    if (strlen(qos_port) == 0)
        qos_do_add_strip_tag = 0;
    else
        qos_do_add_strip_tag = 1;

    if (nvram_match("emf_enable", "1"))
        igmp_snooping_enable = 1;
    else
        igmp_snooping_enable = 0;

    if (igmp_snooping_enable)
    {
        qos_enable = 1;
        qos_do_add_strip_tag = 1;
    }

    if (qos_enable && qos_do_add_strip_tag)
        dma_qos_enable = 1;
    else
        dma_qos_enable = 0;
    return 0;
}   

static int et_rx_strip_bcm_tag(et_info_t *et, struct sk_buff *skb)
{
    unsigned long *pMacHdr;
    unsigned long hdr[3];
    struct ethhdr *eth_hdr;
    int port=0;

    eth_hdr = (struct ethhdr *)&(skb->data[0]);

    /* 2. Check whether BCM tag is added */
    if (ntohs(eth_hdr->h_proto) == ETHER_TYPE_BCM_TAG)
    {
#ifdef _DEBUG_QOS_INPUT
        printk("Before strip (%d): ", skb->len);
        for (i=0; i<skb->len; i++)
            printk("%02X ", skb->data[i]);
        printk("\n");
#endif

        /* Store the src and dst MAC address first */
        pMacHdr = (unsigned long *)&(skb->data[0]);
        hdr[0] = pMacHdr[0];
        hdr[1] = pMacHdr[1];
        hdr[2] = pMacHdr[2];

        /* Check port number and assign priority */
        port = skb->data[SKB_PORT_OFFSET] & 0x0F;
        if (port >=1 && port <= 4)
        {
            /* Set port number to skb->cb[48-11] */
            skb->cb[sizeof(skb->cb)-11] = (5-port);
#ifdef DEBUG
            printk("QoS: rx port#%d\n", 5-port);
#endif
        }

#if defined(__CONFIG_IGMP_SNOOPING__)
        //printk("QoS: rx port#%d, p%d p%d p%d\n", port ,skb->data[29],skb->data[30],skb->data[31]);
        if((skb->data[0]==0x01)&&bcm_robo_snoop_mac_match(&(skb->data[0]))
            && igmp_snooping_enable && is_reg_snooping_enable)
        {//replace BCM tag with vlan ID
            if(port==0)//from wan then send to wireless vlan
            {
                /* Remove 2 byte Broadcom tag */
                skb_pull(skb, 2);
                pMacHdr = (unsigned long *)&(skb->data[0]);
                pMacHdr[0] = hdr[0];
                pMacHdr[1] = hdr[1];
                pMacHdr[2] = hdr[2];
                pMacHdr[3] = 0x1000081;
            }
            else//from lan then send to cpu port
            {
                /* Remove 2 byte Broadcom tag */
                skb_pull(skb, 2);
                pMacHdr = (unsigned long *)&(skb->data[0]);
                pMacHdr[0] = hdr[0];
                pMacHdr[1] = hdr[1];
                pMacHdr[2] = hdr[2];
                pMacHdr[3] = 0x0000081;
            }
        }
        else//not muticast packet
#endif
        {
            /* Remove Broadcom tag */
            skb_pull(skb, BCM_TAG_SIZE);

            pMacHdr = (unsigned long *)&(skb->data[0]);
            pMacHdr[0] = hdr[0];
            pMacHdr[1] = hdr[1];
            pMacHdr[2] = hdr[2];
        }
        /* Remove trailing CRC */
        //skb_trim(skb, skb->len - CRC_SIZE);

#ifdef _DEBUG_QOS_INPUT
        printk("AFTER strip (%d): ", skb->len);
        for (i=0; i<skb->len; i++)
            printk("%02X ", skb->data[i]);
        printk("\n");
#endif        
    }

    
    /* Mark a bit in skb to indicate this is rx from LAN ethernet port */
#if defined(__CONFIG_IGMP_SNOOPING__)
    if (igmp_snooping_enable && is_reg_snooping_enable)
        skb->cb[sizeof(skb->cb)-8]= port;
#endif
#ifdef DEBUG
    printk("QoS: pkt pri = %d port =%d\n", skb->cb[sizeof(skb->cb)-3],skb->cb[sizeof(skb->cb)-8]);
#endif

    return 0;
}

static struct sk_buff * et_tx_add_bcm_tag(et_info_t *et, struct sk_buff **pskb)
{ 
    struct sk_buff *skb = *pskb;
    struct sk_buff *new_skb;
    int i;
    uint32 *old_hdr, *new_hdr;
    uint16 *bcm_tag;
    uint32 *bcm_tag_value;
    int new_tailroom = 0;
    char *crc;

#ifdef _DEBUG_QOS_OUTPUT
    printk("--- Before Insert (%d): ", skb->len);
    for (i=0; i<skb->len; i++)
        printk("%02X ", skb->data[i]);
    printk("\n");
#endif

    if (skb->len < MIN_PACKET_SIZE)
        new_tailroom = MIN_PACKET_SIZE - skb->len;

    if (new_tailroom < CRC_SIZE)
        new_tailroom = CRC_SIZE;

    /* Reallocate skbuff if we are short in headroom or tailroom */
    if (skb_headroom(skb) < BCM_TAG_SIZE || skb_tailroom(skb) < new_tailroom)
    {
        new_skb = skb_copy_expand(skb, BCM_TAG_SIZE, new_tailroom, GFP_ATOMIC);
        if (!new_skb)
        {
            //printk("No room for BRCM tag\n");
            return skb;
        }
        kfree_skb(skb);
        skb = new_skb;
    }

    old_hdr = (uint32 *)&(skb->data[0]);
    new_hdr = (uint32 *)skb_push(skb, BCM_TAG_SIZE);
            
    /* Move the dest and src mac to new location */
    for (i=0; i<3; i++)
        new_hdr[i] = old_hdr[i];

    /* Add Broadcom tag */
    bcm_tag  = (uint16 *)&new_hdr[3];
    *bcm_tag = htons(ETHER_TYPE_BCM_TAG);
    bcm_tag_value  = (uint32 *)&bcm_tag[1];
    *bcm_tag_value = htonl(0x5L);
    
    crc = (char *)skb_put(skb, new_tailroom);

#ifdef _DEBUG_QOS_OUTPUT
    printk("--- AFTER Insert (%d): ", skb->len);
    for (i=0; i<skb->len; i++)
        printk("%02X ", skb->data[i]);
    printk("\n");
#endif
        
    return skb;  
}

#endif      /* INCLUDE_QOS */
